from __future__ import division

from ..optics import OpticalElement
from ..field import Field
from ..propagation import FresnelPropagator

import numpy as np
from scipy.special import gamma, kv

class AtmosphericLayer(OpticalElement):
    '''A single infinitely-thin atmospheric layer.

    This class serves as a base class for all atmospheric layers. Multiply
    atmospheric layers can be combined into an :class:`MultiLayerAtmosphere` which
    provides modelling of scintillation between layers.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the wavefront will be defined.
    Cn_squared : scalar
        The integrated value of Cn^2 for the layer.
    L0 : scalar
        The outer scale of the layer.
    velocity : scalar or array_like
        The velocity of the layer. If a scalar is given, its direction will be
        along x.
    height : scalar
        The height of the atmospheric layer above the ground.

    Attributes
    ----------
    input_grid : Grid
        The grid on which the wavefront will be defined.
    t : scalar
        The current time.
    Cn_squared : scalar
        The integrated value of Cn^2 for the layer.
    L0 : scalar
        The outer scale of the phase structure function.
    velocity : array_like
        The two-dimensional velocity of the layer.
    height : scalar
        The height of the atmospheric layer above the ground.
    '''
    def __init__(self, input_grid, Cn_squared=None, L0=np.inf, velocity=0, height=0):
        self.input_grid = input_grid
        self.Cn_squared = Cn_squared
        self.L0 = L0

        self._velocity = None
        self.velocity = velocity

        self.height = height
        self._t = 0

    def evolve_until(self, t):
        '''Evolve the atmospheric layer until time `t`.

        Parameters
        ----------
        t : scalar
            The time to which to evolve the atmospheric layer.
        '''
        raise NotImplementedError()

    def reset(self):
        '''Reset the phase screen.

        This will create a randomized uncorrelated phase screen.
        '''
        raise NotImplementedError()

    @property
    def t(self):
        '''The current time of the atmospheric layer.
        '''
        return self._t

    @t.setter
    def t(self, t):
        self.evolve_until(t)

    @property
    def Cn_squared(self):  # noqa: N802
        '''The integrated value of Cn^2 for the layer.
        '''
        return self._Cn_squared

    @Cn_squared.setter
    def Cn_squared(self, Cn_squared):  # noqa: N802
        raise NotImplementedError()

    @property
    def outer_scale(self):
        '''The outer scale of the phase structure function.
        '''
        return self._outer_scale

    @outer_scale.setter
    def outer_scale(self, L0):
        raise NotImplementedError()

    @property
    def L0(self):  # noqa: N802
        '''The outer scale of the phase structure function.
        '''
        return self.outer_scale

    @L0.setter
    def L0(self, L0):  # noqa: N802
        self.outer_scale = L0

    @property
    def velocity(self):
        '''The two-dimensional velocity of the layer.
        '''
        return self._velocity

    @velocity.setter
    def velocity(self, velocity):
        if np.isscalar(velocity):
            self._velocity = np.array([velocity, 0])
        else:
            self._velocity = np.array(velocity)

    def phase_for(self, wavelength):
        '''Get the phase screen in radians at a certain wavelength.

        Each atmospheric layer is modelled as an infinitely-thin phase screen.

        Parameters
        ----------
        wavelength : scalar
            The wavelength at which to calculate the phase screen.
        '''
        raise NotImplementedError()

    @property
    def output_grid(self):
        return self.input_grid

    def forward(self, wf):
        wf = wf.copy()
        wf.electric_field *= np.exp(1j * self.phase_for(wf.wavelength))
        return wf

    def backward(self, wf):
        wf = wf.copy()
        wf.electric_field *= np.exp(-1j * self.phase_for(wf.wavelength))
        return wf

class MultiLayerAtmosphere(OpticalElement):
    '''A multi-layer atmospheric model.

    This :class:`OpticalElement` can model turbulence and scintillation effects
    due to atmospheric turbulence by propagating light through a series of
    infinitely-thin atmospheric phase screens at different altitudes. The distance
    between two phase screens can be propagated using Fresnel propagation, or using
    no :class:`Propagator`.

    Parameters
    ----------
    layers : list of AtmosphericLayer objects
        The series of atmospheric layers in this model.
    scintillation : bool
        If True, then the distance between two phase screens is propagated using
        a :class:`FresnelPropagator`. Otherwise, no propagator will be used.
    '''
    def __init__(self, layers, scintillation=False, scintilation=None):
        # Retain backwards compatibility.
        if scintilation is not None:
            import warnings
            warnings.warn('Please use the correct spelling for scintillation.')

            scintillation = scintilation

        self.layers = layers
        self._scintillation = scintillation
        self._t = 0
        self._dirty = True

        self.calculate_propagators()

    def calculate_propagators(self):
        '''Recalculates the list of optical elements used for a propagation.

        This function is called automatically by other functions, but a recalculation
        can be forced by calling it explicitly.
        '''
        heights = np.array([l.height for l in self.layers])
        layer_indices = np.argsort(-heights)

        sorted_heights = heights[layer_indices]
        delta_heights = sorted_heights[:-1] - sorted_heights[1:]
        grid = self.layers[0].input_grid

        if self.scintillation:
            propagators = [FresnelPropagator(grid, h) for h in delta_heights]

        self.elements = []
        for i, j in enumerate(layer_indices):
            self.elements.append(self.layers[j])
            if self.scintillation and i < len(propagators):
                self.elements.append(propagators[i])

        if self.scintillation and sorted_heights[-1] > 0:
            self.elements.append(FresnelPropagator(grid, sorted_heights[-1]))

        self._dirty = False

    def reset(self):
        for l in self.layers:
            l.reset()

    @property
    def layers(self):
        '''A list of :class:`AtmosphericLayer` objects.
        '''
        return self._layers

    @layers.setter
    def layers(self, layers):
        self._layers = layers
        self._dirty = True

    def phase_for(self, wavelength):
        '''Get the unwrapped phase in radians for the atmosphere.

        Parameters
        ----------
        wavelength : scalar
            The wavelength at which to calculate the phase screen.

        Returns
        -------
        Field
            The total unwrapped phase screen.
        '''
        if self.scintillation:
            raise ValueError('Cannot get the unwrapped phase for an atmosphere with scintillation.')

        unwrapped_phases = [layer.phase_for(wavelength) for layer in self.layers]

        return Field(np.sum(unwrapped_phases, axis=0), unwrapped_phases[-1].grid)

    @property
    def scintillation(self):
        '''Whether to include scintillation effects in the propagation.
        '''
        return self._scintillation

    @scintillation.setter
    def scintillation(self, scintillation):
        self._dirty = scintillation != self.scintillation
        self._scintillation = scintillation

    def evolve_until(self, t):
        '''Evolve all atmospheric layers to a time t.

        Parameters
        ----------
        t : scalar
            The time to which to evolve the atmospheric layers.
        '''
        for l in self.layers:
            l.evolve_until(t)
        self._t = t

    @property
    def Cn_squared(self):  # noqa: N802
        '''The total Cn^2 value of the simulated atmosphere.
        '''
        return np.sum([l.Cn_squared for l in self.layers])

    @Cn_squared.setter
    def Cn_squared(self, Cn_squared):  # noqa: N802
        old_Cn_squared = self.Cn_squared
        for l in self.layers:
            l.Cn_squared = l.Cn_squared / old_Cn_squared * Cn_squared

    @property
    def outer_scale(self):
        '''The outer scale of all layers.
        '''
        return self.layers[0].outer_scale

    @outer_scale.setter
    def outer_scale(self, L0):
        for l in self.layers:
            l.outer_scale = L0

    @property
    def t(self):
        '''The current time.
        '''
        return self._t

    @t.setter
    def t(self, t):
        self.evolve_until(t)

    def forward(self, wavefront):
        if self._dirty:
            self.calculate_propagators()

        wf = wavefront.copy()
        for el in self.elements:
            wf = el.forward(wf)
        return wf

    def backward(self, wavefront):
        if self._dirty:
            self.calculate_propagators()

        wf = wavefront.copy()
        for el in reversed(self.elements):
            wf = el.backward(wf)
        return wf

def phase_covariance_von_karman(r0, L0):
    '''Return a Field generator for the phase covariance function for Von Karman turbulence.

    Parameters
    ----------
    r0 : scalar
        The Fried parameter in meters.
    L0 : scalar
        The outer scale in meters.

    Returns
    -------
    Field generator
        The phase covariance Field generator.
    '''
    def func(grid):
        r = grid.as_('polar').r + 1e-10

        a = (L0 / r0)**(5 / 3)
        b = gamma(11 / 6) / (2**(5 / 6) * np.pi**(8 / 3))
        c = (24 / 5 * gamma(6 / 5))**(5 / 6)
        d = (2 * np.pi * r / L0)**(5 / 6)
        e = kv(5 / 6, 2 * np.pi * r / L0)

        return Field(a * b * c * d * e, grid)
    return func

def phase_structure_function_von_karman(r0, L0):
    '''Return a Field generator for the phase structure function for Von Karman turbulence.

    Parameters
    ----------
    r0 : scalar
        The Fried parameter in meters.
    L0 : scalar
        The outer scale in meters.

    Returns
    -------
    Field generator
        The phase structure Field generator.
    '''
    def func(grid):
        r = grid.as_('polar').r + 1e-10

        a = (L0 / r0)**(5 / 3)
        b = 2**(1 / 6) * gamma(11 / 6) / np.pi**(8 / 3)
        c = (24 / 5 * gamma(6 / 5))**(5 / 6)
        d = gamma(5 / 6) / 2**(1 / 6)
        e = (2 * np.pi * r / L0)**(5 / 6)
        f = kv(5 / 6, 2 * np.pi * r / L0)

        return Field(a * b * c * (d - e * f), grid)
    return func

def power_spectral_density_von_karman(r0, L0):
    '''Return a Field generator for the power spectral density function for Von Karman turbulence.

    Parameters
    ----------
    r0 : scalar
        The Fried parameter in meters.
    L0 : scalar
        The outer scale in meters.

    Returns
    -------
    Field generator
        The power spectral density Field generator.
    '''
    def func(grid):
        u = grid.as_('polar').r + 1e-10
        u0 = 2 * np.pi / L0

        res = 0.0229 * ((u**2 + u0**2) / (2 * np.pi)**2)**(-11 / 6.) * r0**(-5 / 3)
        res[u < 1e-9] = 0

        return Field(res, grid)
    return func

def Cn_squared_from_fried_parameter(r0, wavelength=500e-9):  # noqa: N802
    '''Calculate the integrated Cn^2 for a certain Fried parameter.

    Parameters
    ----------
    r0 : scalar
        The Fried parameter in meters.
    wavelength : scalar
        The wavelength in meters at which the Fried parameter is measured (default: 500nm).

    Returns
    -------
    scalar
        The integrated Cn^2 value for the atmosphere.
    '''
    k = 2 * np.pi / wavelength
    return r0**(-5. / 3) / (0.423 * k**2)

def fried_parameter_from_Cn_squared(Cn_squared, wavelength=500e-9):  # noqa: N802
    '''Calculate the Fried parameter from the integrated Cn^2.

    Parameters
    ----------
    r0 : scalar
        The integrated Cn^2 value for the atmosphere.
    wavelength : scalar
        The wavelength in meters at which to calculate the Fried parameter (default: 500nm).

    Returns
    -------
    scalar
        The Fried parameter in meters.
    '''
    k = 2 * np.pi / wavelength
    return (0.423 * Cn_squared * k**2)**(-3. / 5)

def seeing_to_fried_parameter(seeing, wavelength=500e-9):
    r'''Calculate the Fried parameter from the seeing FWHM.

    This uses :math:`\theta = 0.98 \lambda / r_0`, where :math:`\theta`
    is the FWHM of the seeing, :math:`\lambda` is the wavelength and
    :math:`r_0` is the Fried parameter.

    Parameters
    ----------
    seeing : scalar
        The FWHM of the seeing in arcsec.
    wavelength : scalar
        The wavelength in meters at which the seeing is measured (default: 500nm).

    Returns
    -------
    scalar
        The Fried parameter in meters at the specified wavelength.
    '''
    return 0.98 * wavelength / np.radians(seeing / 3600)

def fried_parameter_to_seeing(r0, wavelength=500e-9):
    r'''Calculate the FWHM of the seeing from the Fried parameter.

    This uses :math:`\theta = 0.98 \lambda / r_0`, where :math:`\theta`
    is the FWHM of the seeing, :math:`\lambda` is the wavelength and
    :math:`r_0` is the Fried parameter.

    Parameters
    ----------
    r0 : scalar
        The Fried parameter in meters.
    wavelength : scalar
        The wavelength in meters at which the Fried parameter is measured (default: 500nm).

    Returns
    -------
    scalar
        The FWHM of the seeing in arcsec at the specified wavelength.
    '''
    return np.degrees(0.98 * wavelength / r0) * 3600



__all__ = [
    'Grid',
    'Field',
    'CartesianGrid',
    'PolarGrid',
    'UnstructuredCoords',
    'SeparatedCoords',
    'RegularCoords',
    'field_einsum',
    'field_dot',
    'field_trace',
    'field_inv',
    'field_inverse_tikhonov',
    'field_inverse_truncated',
    'field_inverse_truncated_modal',
    'field_svd',
    'make_field_operation',
    'field_conjugate_transpose',
    'field_transpose',
    'field_determinant',
    'field_adjoint',
    'field_cross',
    'field_kron',
    'make_uniform_grid',
    'make_pupil_grid',
    'make_focal_grid_from_pupil_grid',
    'make_focal_grid',
    'make_hexagonal_grid',
    'make_chebyshev_grid',
    'make_supersampled_grid',
    'make_subsampled_grid',
    'subsample_field',
    'evaluate_supersampled',
    'make_uniform_vector_field',
    'make_uniform_vector_field_generator',
    'is_field',
]

from .grid import *
from .coordinates import *
from .cartesian_grid import *
from .operations import *
from .polar_grid import *
from .spherical_grid import *
from .field import *
from .util import *



import numpy as np

from .chirp_z_transform import ChirpZTransform
from .fourier_transform import FourierTransform, _get_float_and_complex_dtype
from ..field import Field

class ZoomFastFourierTransform(FourierTransform):
    '''A Zoom Fast Fourier transform (ZoomFFT) object.

    This Fourier transform is a specialization of the Chirp Z-transform. It requires
    both the input and output grid to be regularly spaced in Cartesian coordinates. However,
    contrary to the Fast Fourier Transform (FFT), the spacing can be arbitrary and a small
    region of Fourier space can be efficiently evaluated.

    The ZoomFFT is asymptotically faster than a Matrix Fourier Transform (MFT) in cases where
    both input and output grids are large, typically at 1k x 1k or bigger in each grid. It also
    supports arbitrary dimenionality of the input and output grids.

    Parameters
    ----------
    input_grid : Grid
        The grid that is expected for the input field.
    output_grid : Grid
        The grid that is produced by the Fourier transform.

    Raises
    ------
    ValueError
        If the input grid is not separated in Cartesian coordinates, if it's not one- or two-
        dimensional, or if the output grid has a different dimension than the input grid.
    '''
    def __init__(self, input_grid, output_grid):
        if not input_grid.is_regular or not input_grid.is_('cartesian'):
            raise ValueError('The input grid should be regularly spaced in Cartesian coordinates.')
        if not output_grid.is_regular or not output_grid.is_('cartesian'):
            raise ValueError('The output grid should be regularly spaced in Cartesian coordinates.')
        if input_grid.ndim != output_grid.ndim:
            raise ValueError('The input_grid must have the same dimensions as the output_grid.')

        self.input_grid = input_grid
        self.output_grid = output_grid

        self._current_dtype = None

    def _compute_shifts_and_weights(self, dtype):
        float_dtype, complex_dtype = _get_float_and_complex_dtype(dtype)

        if complex_dtype != self._current_dtype:
            w = np.exp(-1j * self.output_grid.delta * self.input_grid.delta)
            a = np.exp(1j * self.output_grid.zero * self.input_grid.delta)

            inv_w = np.exp(1j * self.input_grid.delta * self.output_grid.delta)
            inv_a = np.exp(-1j * self.input_grid.zero * self.output_grid.delta)

            self.czts = [ChirpZTransform(n, m, ww, aa) for n, m, ww, aa in zip(self.input_grid.dims, self.output_grid.dims, w, a)]
            self.inv_czts = [ChirpZTransform(n, m, ww, aa) for n, m, ww, aa in zip(self.output_grid.dims, self.input_grid.dims, inv_w, inv_a)]

            self.shifts = [np.exp(-1j * x * x0) for x, x0 in zip(self.output_grid.separated_coords, self.input_grid.zero)]
            self.inv_shifts = [np.exp(1j * x * x0) for x, x0 in zip(self.input_grid.separated_coords, self.output_grid.zero)]

            self.shifts = [s.astype(complex_dtype, copy=False) for s in self.shifts]
            self.inv_shifts = [s.astype(complex_dtype, copy=False) for s in self.inv_shifts]

            self.input_weights = self.input_grid.weights.astype(float_dtype)
            self.output_weights = (self.output_grid.weights / (2 * np.pi)**self.output_grid.ndim).astype(float_dtype)

            self._current_dtype = complex_dtype

    def forward(self, field):
        '''Returns the forward Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to Fourier transform.

        Returns
        -------
        Field
            The Fourier transform of the field.
        '''
        self._compute_shifts_and_weights(field.dtype)

        f = (field * self.input_weights).shaped

        for i, (czt, shift) in enumerate(zip(self.czts, self.shifts)):
            f = np.moveaxis(f, -i, 0)
            f = czt(f) * shift
            f = np.moveaxis(f, -i, 0)

        shape = tuple(field.tensor_shape) + (-1,)

        return Field(f.reshape(shape), self.output_grid)

    def backward(self, field):
        '''Returns the inverse Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to inverse Fourier transform.

        Returns
        -------
        Field
            The inverse Fourier transform of the field.
        '''
        self._compute_shifts_and_weights(field.dtype)

        f = (field * self.output_weights).shaped

        for i, (czt, shift) in enumerate(zip(self.inv_czts, self.inv_shifts)):
            f = np.moveaxis(f, -i, 0)
            f = czt(f) * shift
            f = np.moveaxis(f, -i, 0)

        shape = tuple(field.tensor_shape) + (-1,)

        return Field(f.reshape(shape), self.input_grid)



import numpy as np

from .optical_element import OpticalElement

class SimpleVibration(OpticalElement):
    def __init__(self, mode, amplitude, frequency, phase_0=0):
        self.mode = mode
        self.amplitude = amplitude
        self.frequency = frequency
        self.phase_0 = phase_0
        self.t = 0

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, frequency):
        delta_phase = 2 * np.pi * (self.frequency + self.frequency)
        self.phase_0 += delta_phase
        self._frequency = frequency

    @property
    def phase(self):
        return 2 * np.pi * self.frequency * self.t + self.phase_0

    def forward(self, wavefront):
        wf = wavefront.copy()

        wf.electric_field *= np.exp(1j * (self.mode * self.amplitude / wf.wavelength * np.sin(self.phase)))
        return wf

    def backward(self, wavefront):
        wf = wavefront.copy()

        wf.electric_field *= np.exp(-1j * (self.mode * self.amplitude / wf.wavelength * np.sin(self.phase)))
        return wf



import numpy as np
from math import sqrt
from scipy.special import eval_hermite

def index_to_hermite(i):
    '''Converts a one-dimensional mode index to the two-dimensional mode index.

    The one-dimensional mode index is converted to a two-dimensional index.
    The two-dimensional index determines the order in the x and y directions.

    Parameters
    ----------
    i : int
        The one-dimensional index.

    Returns
    -------
    n : int
        The x order.
    m : int
        The y order.
    '''

    # Get the max order
    o = int((np.sqrt(8 * i + 1) - 1) / 2)

    # Get the y order
    m = int(i - o * (o + 1) / 2)

    # Get the x order
    n = o - m

    return n, m

def gaussian_hermite(n, m, mode_field_diameter=1, grid=None):
    r'''Creates a Gaussian-Hermite mode.

    This function evaluates a (n, m) order Gaussian-Hermite mode on a grid.
    The definition of the modes are the following,

    .. math::
        \exp{\left(-\frac{r^2}{w_0^2}\right)} H_n\left(\sqrt{2}\frac{x}{w_0}\right) H_m\left(\sqrt{2}\frac{y}{w_0}\right).

    Here :math:`w_0` is the mode_field_radius, which is :math:`\mathrm{MFD}/2`. This defintion follows
    the Physicists definition of the Hermite polynomials. The modes are numerical normalized to have a
    total power of 1.

    More details on the Hermite Polynomials can be found on: http://mathworld.wolfram.com/HermitePolynomial.html

    Parameters
    ----------
    n : int
        The x order.
    m : int
        The y order.
    mode_field_diameter : scalar
        The mode field diameter of the Gaussian-Laguerre mode.
    grid : Grid
        The grid on which to evaluate the mode.

    Returns
    -------
    Field
        The evaluated mode.
    '''
    from ..field import Field

    if grid.is_separated and grid.is_('cartesian'):
        x = grid.x
        y = grid.y
        r2 = (x**2 + y**2) / (mode_field_diameter / 2)**2
    else:
        x, y = grid.as_('cartesian').coords
        r2 = (2 * grid.r / mode_field_diameter)**2

    # Calculate the mode.
    hg = np.exp(-r2) * eval_hermite(n, 2 * sqrt(2) * x / mode_field_diameter) * eval_hermite(m, 2 * sqrt(2) * y / mode_field_diameter)

    # Numerically normalize the mode
    hg /= np.sqrt(np.sum(np.abs(hg)**2 * grid.weights))

    return Field(hg, grid)

def gaussian_hermite_index(i, mode_field_diameter=1, grid=None):
    '''Creates a Gaussian-Hermite mode.

    This function evaluates the i-th order Gaussian-Hermite mode on a grid.

    Parameters
    ----------
    i : int
        The one-dimensional mode index.
    mode_field_diameter : scalar
        The mode field diameter of the Gaussian-Hermite mode.
    grid : Grid
        The grid on which to evaluate the Gaussian-Hermite mode.

    Returns
    -------
    Field
        The evaluated Gaussian-Hermite mode.
    '''

    n, m = index_to_hermite(i)
    return gaussian_hermite(n, m, mode_field_diameter, grid)

def make_gaussian_hermite_basis(grid, num_modes, mode_field_diameter, starting_mode=0):
    '''Creates a Gaussian-Hermite mode basis.

    This function evaluates the starting_mode to num_modes + starting_modes Gaussian-Hermite modes.
    And returns a ModeBasis made out of these Gaussian-Hermite modes.

    Parameters
    ----------
    grid : Grid
        The grid on which to evaluate the Gaussian-Hermite mode.
    num_modes : int
        The number of modes to create.
    mode_field_diameter : scalar
        The mode field diameter of the Gaussian-Hermite mode.
    starting_mode : int
        The starting point of the mode indices.

    Returns
    -------
    ModeBasis
        The evaluated Gaussian-Hermite modes.
    '''
    from .mode_basis import ModeBasis

    modes = [gaussian_hermite_index(i, mode_field_diameter, grid) for i in range(starting_mode, starting_mode + num_modes)]
    return ModeBasis(modes, grid)



import numpy as np
import inspect
import collections
import itertools

class OpticalElement(object):
    '''Base class for all optical elements.

    This class can propagate a :class:`Wavefront` through an optical element
    (or free space), therefore modifying it. Any OpticalElement should be agnostic
    of the grid and wavelength of the wavefront. If it's not, you can use the
    :func:`make_agnostic_optical_element` decorator to create an agnostic optical
    element out of a gnostic one.
    '''
    def __call__(self, wavefront):
        '''Propagate a wavefront forward through the optical element.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        return self.forward(wavefront)

    def forward(self, wavefront):
        '''Propagate a wavefront forward through the optical element.

        This will be implemented by the derived class.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        raise NotImplementedError()

    def backward(self, wavefront):
        '''Propagate a wavefront backward through the optical element.

        A backward propagation is a literal backward propagation through
        the element. It is not (in general) the inverse of the forward
        propagation, except in cases where energy is conserved.

        This function will be implemented by the derived class.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        raise NotImplementedError()

    def get_transformation_matrix_forward(self, wavelength=1):
        '''Calculate the linear transformation matrix that corresponds
        to a forward propagation.

        The matrix is defined such that `E_out = M.dot(E_in)`, where `E_out` and
        `E_in` are the electric fields of the output and input respectively, and
        `M` is the transformation matrix returned by this function.

        :: warning::
            Depending on the chosen resolution, this matrix can be extremely large and
            use extremely large amounts of memory. For example, a Fresnel propagation on
            a 128x128 grid will already create a matrix of 2.1GB. Use with caution.

        This function will be implemented by the derived class.

        Parameters
        ----------
        wavelength : scalar
            The wavelength for which the transformation matrix will be calculated.

        Returns
        -------
        ndarray
            The full propagation matrix.
        '''
        raise NotImplementedError()

    def get_transformation_matrix_backward(self, wavelength=1):
        '''Calculate the linear transformation matrix that corresponds
        to a backward propagation.

        The matrix is defined such that `E_in = M.dot(E_out)`, where `E_out` and
        `E_in` are the electric fields of the output and input plane respectively, and
        `M` is the transformation matrix returned by this function.

        A backward propagation is a literal backward propagation through
        the element. It is not (in general) the inverse of the forward
        propagation, except in cases where energy is conserved.

        :: warning::
            Depending on the chosen resolution, this matrix can be extremely large and
            use extremely large amounts of memory. For example, a Fresnel propagation on
            a 128x128 grid will already create a matrix of 2.1GB. Use with caution.

        This function will be implemented by the derived class.

        Parameters
        ----------
        wavelength : scalar
            The wavelength for which the transformation matrix will be calculated.

        Returns
        -------
        ndarray
            The full propagation matrix.
        '''
        raise NotImplementedError()

    def get_instance(self, input_grid=None, output_grid=None, wavelength=None):
        '''Return an OpticalElement that can handle wavefronts with input_grid and wavelength.

        While any OpticalElement should in theory be able to handle all grids and wavelengths,
        this function is added to simplify the interface for those who don't. It allows the user
        to always access properties of an OpticalElement, evaluated for a specific input_grid
        and wavelength.

        The user needs to supply at least an input grid or an output grid, and a wavelength. If
        this is not done, a ValueError will be raised.

        Parameters
        ----------
        input_grid : Grid
            The grid on which the input wavefront is defined.
        output_grid : Grid or None
            The grid on which the output wavefront is defined.
        wavelength : scalar
            The wavelength on which the wavefront is defined.

        Returns
        -------
        OpticalElement
            An optical element that can handle wavefront with the specified input grid and wavelength.
        '''
        return self

class EmptyOpticalElement(OpticalElement):
    '''An empty optical element.

    This optical element doesn't modify the wavefront at all. This can be used as a replacement
    for optical elements. For example, when you don't want to use a coronagraph for code that expects
    an optical element as a coronagraph, you can pass an instance of this class to effectively do nothing.
    '''
    def forward(self, wavefront):
        '''Propagate the wavefront forward through the empty optical element.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        return wavefront

    def backward(self, wavefront):
        '''Propagate the wavefront backward through the empty optical element.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        return wavefront

    def get_transformation_matrix_forward(self, wavelength=1):
        '''Calculate the backward linear transformation matrix for the empty optical element.

        Parameters
        ----------
        wavelength : scalar
            The wavelength for which the transformation matrix will be calculated.

        Returns
        -------
        ndarray
            The full propagation matrix.
        '''
        return np.array(1)

    def get_transformation_matrix_backward(self, wavelength=1):
        '''Calculate the backward linear transformation matrix for the empty optical element.

        Parameters
        ----------
        wavelength : scalar
            The wavelength for which the transformation matrix will be calculated.

        Returns
        -------
        ndarray
            The full propagation matrix.
        '''
        return np.array(1)

class InstanceData(object):
    '''An object that stores data related to a single instance of input_grid, output_grid and wavelength.

    Parameters
    ----------
    input_grid : Grid or None
        The input grid.
    output_grid : Grid or None
        The output grid.
    wavelength : scalar or None
        The wavelength.
    '''
    def __init__(self, input_grid, output_grid, wavelength):
        self.input_grid = input_grid
        self.output_grid = output_grid
        self.wavelength = wavelength

def _get_function_parameters(func):
    '''Get the names of the parameters for a function for both Python2 and Python3.

    Parameters
    ----------
    func : function
        The function to get the parameter names for.

    Returns
    -------
    list of str
        The list with the name for each of the parameters of the given function.
    '''
    if hasattr(inspect, 'signature'):
        # Python 3
        return list(inspect.signature(func).parameters.keys())
    else:
        # Python 2
        return inspect.getargspec(func).args

INPUT_GRID_DEPENDENT = 1
OUTPUT_GRID_DEPENDENT = 2
WAVELENGTH_DEPENDENT = 4

class AgnosticOpticalElement(OpticalElement):
    '''Base class for optical elements that require additional processing/caching for supporting different grids or wavelengths.

    This class is mant to simplify the creation of agnostic optical elements. When you have an
    optical element that explicitly needs an input grid and/or wavelength on initialization, you can
    use this class to simplify making it accept all input/output grids and/or wavelengths.

    Instances are created by a function `make_instance` that derived classes should use to evaluate properties
    on the given input_grid, output_grid and wavelength. These properties are stored in an internal cache, and
    are reused for as long as they reside in the cache. Any function using forward or backward propagation will
    have access to the cached instance data. All caching and retrieving of cached data is made invisible to the
    user.

    As instanced data can take a lot of memory, at most `max_in_cache` instances will be held in the cache at
    the time. If an additional instance is requested, the oldest instance will be thrown away, and the new
    instance will be put in the cache instead.

    Parameters
    ----------
    grid_dependent : boolean
        If the instances should be separated by grid. Separate instances will be made if their grids change
        between invocations.
    wavelength_dependent : boolean
        If the instances should be separated by grid.Separate instances will be made if wavelength changes
        between invocations.
    max_in_cache : int
        The maximum size of the internal cache for optical elements. Reduce this if the cache is using
        too much memory, increase if there are a lot of cache misses.
    '''
    def __init__(self, grid_dependent=True, wavelength_dependent=True, max_in_cache=11):
        self._grid_dependent = grid_dependent
        self._wavelength_dependent = wavelength_dependent
        self._max_in_cache = max_in_cache

        self.clear_cache()

    def _get_cache_keys(self, input_grid, output_grid, wavelength):
        '''Get all keys for the instance cache.

        Parameters
        ----------
        input_grid : Grid or None
            The input grid.
        output_grid : Grid or None
            The output grid.
        wavelength : scalar or None
            The wavelength.

        Returns
        -------
        list
            The list of cache keys.
        '''
        key_parts = []

        if self._grid_dependent:
            if input_grid is None:
                if output_grid is None:
                    raise ValueError('Grid dependent, but no grids are given for lookup.')

                key_parts.append([(None, hash(output_grid))])
            elif output_grid is None:
                key_parts.append([(hash(input_grid), None)])
            else:
                key_parts.append([(hash(input_grid), hash(output_grid)), (hash(input_grid), None), (None, hash(output_grid))])
        else:
            key_parts.append([(None, None)])

        if self._wavelength_dependent:
            if wavelength is None:
                raise ValueError('Wavelength dependent, but no wavelength is given for lookup.')

            wavelength_key = int(np.round(np.log(wavelength) / np.log(1 + 1e-9)))
            key_parts.append([(wavelength_key,)])
        else:
            key_parts.append([(None,)])

        keys = []
        for parts in itertools.product(*key_parts):
            key = ()
            for p in parts:
                key += p
            keys.append(key)

        return keys

    def clear_cache(self):
        '''Clear the instance cache.

        This function should be called if agnostic data, that was used
        to create instance data, was changed by the user. Clearing the
        cache ensures that the propagations are always performed using
        up-to-date arguments.
        '''
        self._instance_data_cache = collections.OrderedDict()
        self._num_in_cache = 0

    def _add_to_cache(self, instance_data, cache_keys=None):
        if self._num_in_cache == self._max_in_cache:
            # Remove last added item
            key, value = self._instance_data_cache.popitem(False)

            # Remove all copies of that item as well
            old_cache_keys = self._get_cache_keys(value.input_grid, value.output_grid, value.wavelength)
            for i in range(len(old_cache_keys) - 1):
                self._instance_data_cache.popitem(False)

            # Update number of instance data items
            self._num_in_cache -= 1

        # Calculate cache keys
        if cache_keys is None:
            cache_keys = self._get_cache_keys(instance_data.input_grid, instance_data.output_grid, instance_data.wavelength)

        # Add instance data under each of these keys
        for cache_key in cache_keys:
            self._instance_data_cache[cache_key] = instance_data

        # Update number of instance data items
        self._num_in_cache += 1

    def _get_parameter_signature(self, parameter):
        '''Guess the signature of a given parameter.

        A parameter can be a function of (input_grid, output_grid, wavelength)
        or a subset of this. This function will try to guess what (sub)set should
        be used to evaluate this parameter.

        Parameters
        ----------
        parameter : anything
            The parameter for which to guess the signature.

        Returns
        -------
        int
            A bit-map indicating the best-guess signature.
        '''
        if not callable(parameter):
            # Parameter is not callable, so no evaluation can be done.
            return 0

        param_parameters = _get_function_parameters(parameter)

        if len(param_parameters) == 1:
            if self._grid_dependent and self._wavelength_dependent:
                # Need to choose between the two
                if 'grid' in param_parameters[0]:
                    return INPUT_GRID_DEPENDENT
                elif 'lam' in param_parameters[0] or 'wave' in param_parameters[0] or 'wvl' in param_parameters[0]:
                    return WAVELENGTH_DEPENDENT
                return WAVELENGTH_DEPENDENT
            elif self._grid_dependent:
                # Only grid dependent
                return INPUT_GRID_DEPENDENT
            elif self._wavelength_dependent:
                # Only wavelength dependent
                return WAVELENGTH_DEPENDENT
            else:
                return 0
        elif len(param_parameters) == 2:
            if not self._wavelength_dependent:
                return INPUT_GRID_DEPENDENT | OUTPUT_GRID_DEPENDENT
            else:
                return INPUT_GRID_DEPENDENT | WAVELENGTH_DEPENDENT
        elif len(param_parameters) == 3:
            return INPUT_GRID_DEPENDENT | OUTPUT_GRID_DEPENDENT | WAVELENGTH_DEPENDENT
        else:
            return 0

    def evaluate_parameter(self, parameter, input_grid, output_grid, wavelength):
        '''Evaluate the parameter as function of (input_grid, output_grid, wavelength).

        The parameter can be a function of all or a subset of these parameters. This function
        will try to guess the function signature and attempt evaluation of the given function.

        Parameters
        ----------
        parameter : anything
            The parameter to evaluate.
        input_grid : Grid or None
            The input grid.
        output_grid : Grid or None
            The output grid.
        wavelength : scalar or None
            The wavelength.

        Returns
        -------
        any type
            The evaluated parameter.

        Raises
        ------
        RuntimeError
            If the function could not be evaluated using the best-guess signature.
        '''
        signature = self._get_parameter_signature(parameter)

        if not signature:
            return parameter

        args = []
        if signature & INPUT_GRID_DEPENDENT:
            if input_grid is None:
                input_grid = self.get_input_grid(output_grid, wavelength)

            args.append(input_grid)
        if signature & OUTPUT_GRID_DEPENDENT:
            if output_grid is None:
                output_grid = self.get_output_grid(input_grid, wavelength)

            args.append(output_grid)
        if signature & WAVELENGTH_DEPENDENT:
            args.append(wavelength)

        try:
            return parameter(*args)
        except Exception:
            raise RuntimeError('Parameter could not be evaluated.')

    def construct_function(self, function, *args, **kwargs):
        '''Construct a function based on the given function and arguments.

        The arguments can be (input_grid, output_grid, wavelength) or any subset of this.
        The returned function has parameters which encompass all given parameters.

        This function is especially usefull for creating properties that depend on
        parameters that depend on input_grid, output_grid and/or wavelength.

        Parameters
        ----------
        function : function
            The function on which to base the returned function.
        *args : anything
            The arguments for the given function.
        **kwargs : anything
            The keyword arguments for the given function.

        Returns
        -------
        function
            The constructed function.

        Raises
        ------
        RuntimeError
            If the signature of one of the parameter is not recognized.
        '''
        # Evaluate function with parameters that can be grid and/or wavelength dependent
        # This returns a function with arguments chosen to fit whether the whole thing is grid and/or wavelength dependent.
        signature = 0

        for arg in args:
            signature |= self._get_parameter_signature(arg)

        for kwarg in kwargs.values():
            signature |= self._get_parameter_signature(kwarg)

        # Check original function for direct parameter dependencies
        function_params = _get_function_parameters(function)

        if 'input_grid' in function_params:
            kwargs['input_grid'] = lambda input_grid: input_grid
            signature |= INPUT_GRID_DEPENDENT

        if 'output_grid' in function_params:
            kwargs['output_grid'] = lambda output_grid: output_grid
            signature |= OUTPUT_GRID_DEPENDENT

        if 'wavelength' in function_params:
            kwargs['wavelength'] = lambda wavelength: wavelength
            signature |= WAVELENGTH_DEPENDENT

        # If parameters are not a function of anything, evaluate the function now and return the result
        if signature == 0:
            return function(*args, **kwargs)

        # Put all args in kwargs, but ignore input_grid, output_grid and wavelength
        index = 0
        for param_name in function_params:
            if param_name in ['input_grid', 'output_grid', 'wavelength']:
                continue

            kwargs[param_name] = args[index]
            index += 1

            if len(args) == index:
                break

        # Create the returned function
        if signature == (INPUT_GRID_DEPENDENT | OUTPUT_GRID_DEPENDENT | WAVELENGTH_DEPENDENT):
            def func(input_grid, output_grid, wavelength):
                evaluated_kwargs = {}
                for key, val in kwargs.items():
                    evaluated_kwargs[key] = self.evaluate_parameter(val, input_grid, output_grid, wavelength)

                return function(**evaluated_kwargs)
        elif signature == (INPUT_GRID_DEPENDENT | OUTPUT_GRID_DEPENDENT):
            def func(input_grid, output_grid):
                evaluated_kwargs = {}
                for key, val in kwargs.items():
                    evaluated_kwargs[key] = self.evaluate_parameter(val, input_grid, output_grid, None)

                return function(**evaluated_kwargs)
        elif signature == (INPUT_GRID_DEPENDENT | WAVELENGTH_DEPENDENT):
            def func(input_grid, wavelength):
                evaluated_kwargs = {}
                for key, val in kwargs.items():
                    evaluated_kwargs[key] = self.evaluate_parameter(val, input_grid, None, wavelength)

                return function(**evaluated_kwargs)
        elif signature == INPUT_GRID_DEPENDENT:
            def func(input_grid):
                evaluated_kwargs = {}
                for key, val in kwargs.items():
                    evaluated_kwargs[key] = self.evaluate_parameter(val, input_grid, None, None)

                return function(**evaluated_kwargs)
        elif signature == WAVELENGTH_DEPENDENT:
            def func(wavelength):
                evaluated_kwargs = {}
                for key, val in kwargs.items():
                    evaluated_kwargs[key] = self.evaluate_parameter(val, None, None, wavelength)

                return function(**evaluated_kwargs)
        else:
            raise RuntimeError('Signature %d was not recognized.' % signature)

        return func

    def get_instance_data(self, input_grid, output_grid, wavelength):
        '''Get the InstanceData object corresponding to the given grids and wavelength.

        If no InstanceData can be found in the internal cache, a new instance will be
        constructed.

        Parameters
        ----------
        input_grid : Grid or None
            The input grid.
        output_grid : Grid or None
            The output grid.
        wavelength : scalar or None
            The wavelength.
        '''
        cache_keys = self._get_cache_keys(input_grid, output_grid, wavelength)

        for cache_key in cache_keys:
            if cache_key in self._instance_data_cache:
                instance_data = self._instance_data_cache[cache_key]
                break
        else:
            # Try to guess input and output grid.
            if input_grid is None:
                input_grid = self.get_input_grid(output_grid, wavelength)

            if output_grid is None:
                output_grid = self.get_output_grid(input_grid, wavelength)

            # Recalculate cache keys and try again.
            cache_keys = self._get_cache_keys(input_grid, output_grid, wavelength)

            for cache_key in cache_keys:
                if cache_key in self._instance_data_cache:
                    instance_data = self._instance_data_cache[cache_key]
                    break
            else:
                # Item does not yet exist. Create instanceData element
                instance_data = InstanceData(input_grid, output_grid, wavelength)
                self.make_instance(instance_data, input_grid, output_grid, wavelength)

                # Add instance data to cache.
                self._add_to_cache(instance_data, cache_keys)

        return instance_data

    def make_instance(self, instance_data, input_grid, output_grid, wavelength):
        '''Make an instance for this specific input_grid, output_grid, wavelength.

        This function is intended to be implemented by a derived class. Any properties
        evaluated for the instance can be stored into the instance_data object which
        is stored in the internal cache.

        Parameters
        ----------
        instance_data : InstanceData
            An object storing all data for this instance. This object can be modified this function.
        input_grid : Grid or None
            The input grid.
        output_grid : Grid or None
            The output grid.
        wavelength : scalar or None
            The wavelength.
        '''
        pass

    def get_input_grid(self, output_grid, wavelength):
        '''Calculate a best guess for the input grid given an output grid and wavelength.

        This function is intended to be implemented by a derived class.

        Parameters
        ----------
        output_grid : Grid
            The output grid.
        wavelength : scalar
            The wavelength.

        Returns
        -------
        Grid
            The best-guess input grid based on the given output grid and wavelength.
        '''
        return None

    def get_output_grid(self, input_grid, wavelength):
        '''Calculate a best guess for the output grid given an input grid and wavelength.

        This function is intended to be implemented by a derived class.

        Parameters
        ----------
        input_grid : Grid
            The input grid.
        wavelength : scalar
            The wavelength.

        Returns
        -------
        Grid
            The best-guess output grid based on the given input grid and wavelength.
        '''

        return None

    def __getattr__(self, name):
        '''A redirect for instance data.

        Any attribute that exists only in an instance can be accessed by this redirect.

        Parameters
        ----------
        name : str
            The name of the requested attribute.

        Returns
        -------
        function
            The attribute that can be evaluated on a specific input, output grid and wavelength.
        '''
        def attribute(input_grid=None, output_grid=None, wavelength=None):
            instance_data = self.get_instance_data(input_grid, output_grid, wavelength)

            return getattr(instance_data, name)
        return attribute

    def __getstate__(self):
        '''Get the state of the optical element for pickle.

        Returns
        -------
        dict
            All contained variables.
        '''
        return self.__dict__

    def __setstate__(self, state):
        '''Set the state of the optical element for pickle.

        Parameters
        ----------
        state : dict
            The state of an optical element returned by a __getstate__().
        '''
        self.__dict__ = state

def make_agnostic_forward(forward):
    '''A decorator for a forward function on an AgnosticOpticalElement.

    Any derived class should use this decorator on any forward-type function. This
    allows the function to have access to instance data, in addition to agnostic properties.

    Parameters
    ----------
    forward : function
        The modified forward-type function.

    Returns
    -------
    function
        The new forward-type function that calls the old function with added instance data.
    '''
    def res(self, wavefront, *args, **kwargs):
        # Look up instance data
        instance_data = self.get_instance_data(wavefront.grid, None, wavefront.wavelength)

        return forward(self, instance_data, wavefront, *args, **kwargs)
    return res

def make_agnostic_backward(backward):
    '''A decorator for a backward function on an AgnosticOpticalElement.

    Any derived class should use this decorator on any backward-type function. This
    allows the function to have access to instance data, in addition to agnostic properties.

    Parameters
    ----------
    backward : function
        The modified backward-type function.

    Returns
    -------
    function
        The new backward-type function that calls the old function with added instance data.
    '''
    def res(self, wavefront, *args, **kwargs):
        # Look up instance data
        instance_data = self.get_instance_data(None, wavefront.grid, wavefront.wavelength)

        return backward(self, instance_data, wavefront, *args, **kwargs)
    return res

def make_agnostic_optical_element(grid_dependent_arguments=None, wavelength_dependent_arguments=None, num_in_cache=50):  # pragma: no cover
    '''Create an optical element that is agnostic to input_grid or wavelength from one that is not.

    This decorator is meant to simplify the creation of agnostic optical elements. When you have an
    optical element that explicitly needs an input grid and/or wavelength on initialization, you can
    use this decorator to modify it to make it accept all input grids and/or wavelengths.

    All parameters pass to the __init__() of the returned class will be attempted to be evaluated on
    an input grid or wavelength (depending on if the parameter name appears in `grid_dependent_arguments`
    or `wavelength_dependent_arguments`). This evaluation is done by calling the argument with either
    the input_grid or wavelength, before passing it to the initializer of the optical element provided
    by the user. When an argument can be both dependent on input grid and wavelength, you can pass a
    function with double arguments. This will be evaluated as `arg(input_grid, wavelength)`. If the
    argument only has a single parameter, this function will make a guess on whether it is input_grid
    or wavelength dependent and try both possibilities.

    Parameters
    ----------
    grid_dependent_arguments : list of strings or None
        A list of all names of parameters that could vary with input_grid. These parameters will be
        evaluated on the grid if they are callable. If this is None, this indicates that no parameters
        can depend on input_grid, and that the optical element was already input_grid agnostic.
    wavelength_dependent_arguments : list of strings or None
        A list of all names of parameters that could vary with wavelength. These parameters will be
        evaluated at the wavelength if they are callable. If this is None, this indicates that no parameters
        can depend on wavelength, and that the optical element was already wavelength agnostic.
    num_in_cache : int
        The maximum size of the internal cache for optical elements. Reduce this if the cache is using
        too much memory, increase if there are a lot of cache misses.

    Returns
    -------
    class
        The new optical element class that is now agnostic to input grid and wavelength.

    Raises
    ------
    RuntimeError
        If one of the argument gave an error during evaluation.
    '''
    import warnings
    warnings.warn('This function will be removed in a future release. Please switch to a AgnosticOpticalElement class instead.', DeprecationWarning)

    if grid_dependent_arguments is None:
        grid_dependent_arguments = []

    if wavelength_dependent_arguments is None:
        wavelength_dependent_arguments = []

    def decorator(optical_element_class):
        gnostic_param_names = _get_function_parameters(optical_element_class.__init__)[1:]

        grid_dependent = grid_dependent_arguments or 'input_grid' in gnostic_param_names
        wavelength_dependent = wavelength_dependent_arguments or 'wavelength' in gnostic_param_names

        class AgnosticOpticalElement(OpticalElement):
            def __init__(self, *args, **kwargs):
                self._cache = collections.OrderedDict()

                self._parameters = dict(zip(gnostic_param_names, args))
                self._parameters.update(kwargs)

            def get_instance(self, input_grid=None, output_grid=None, wavelength=None):
                if grid_dependent and ((input_grid is None) == (output_grid is None)):
                    raise ValueError('You need to supply either an input or output grid.')

                if wavelength_dependent and (wavelength is None):
                    raise ValueError('You need to supply a wavelength.')

                # Get cache key
                cache_key = ()

                if grid_dependent:
                    if input_grid is not None:
                        cache_key += ('input', input_grid)
                    else:
                        cache_key += ('output', output_grid)

                if wavelength_dependent:
                    # Use approximate wavelength as a key (match if within 1e-9 relatively).
                    wavelength_key = int(np.round(np.log(wavelength) / np.log(1 + 1e-9)))
                    cache_key += (wavelength_key, )

                # Is there an element in the cache.
                if cache_key in self._cache:
                    return self._cache[cache_key]

                if output_grid is not None:
                    # If we supplied an output grid, it needs to be listed into the cache, as we
                    # cannot initialize/create an optical element from an output grid.
                    raise RuntimeError('Output grid is not known. Perform a forward propagation first before backwards propagation on the same grid.')

                # If the cache is full, remove the oldest element to make room for a new one.
                if len(self._cache) == 2 * num_in_cache:
                    self._cache.popitem(False)
                    self._cache.popitem(False)

                # Create a new element.
                element_parameters = dict(self._parameters)

                if 'input_grid' in gnostic_param_names:
                    element_parameters['input_grid'] = input_grid
                if 'wavelength' in gnostic_param_names:
                    element_parameters['wavelength'] = wavelength

                # Evaluate grid dependent arguments (including double-dependent arguments)
                for param_name in grid_dependent_arguments:
                    if not callable(element_parameters[param_name]):
                        # Argument is not callable, so no evaluation can be done.
                        continue

                    if param_name in wavelength_dependent_arguments:
                        # Argument can be a function of either or both. Check between either or both first.
                        param_parameters = _get_function_parameters(element_parameters[param_name])
                        if len(param_parameters) == 1:
                            # Argument is a function of either, but we do not know which. Look if we have clues:
                            function_of = 'input_grid'
                            if 'grid' in param_parameters[0]:
                                function_of = 'input_grid'
                            elif 'lam' in param_parameters[0] or 'wave' in param_parameters[0] or 'wvl' in param_parameters[0]:
                                function_of = 'wavelength'

                            # Try first choice
                            try:
                                if function_of == 'input_grid':
                                    res = element_parameters[param_name](input_grid)
                                else:
                                    res = element_parameters[param_name](wavelength)
                            except Exception:
                                # Function evaluation failed. Try other one:
                                try:
                                    if function_of == 'input_grid':
                                        res = element_parameters[param_name](wavelength)
                                    else:
                                        res = element_parameters[param_name](input_grid)
                                except Exception:
                                    # Function evaluation failed again. Raise exception.
                                    raise RuntimeError('The argument %s can not be evaluated.' % param_name)

                            element_parameters[param_name] = res
                        else:
                            # Argument is a function of both.
                            try:
                                res = element_parameters[param_name](input_grid, wavelength)
                            except Exception:
                                # Function evaluation failed again. Raise exception.
                                raise RuntimeError('The argument %s can not be evaluated.' % param_name)
                    else:
                        # Argument is a function of input_grid.
                        try:
                            element_parameters[param_name] = element_parameters[param_name](input_grid)
                        except Exception:
                            # Function evaluation failed. Raise exception.
                            raise RuntimeError('The argument %s can not be evaluated.' % param_name)

                # Evaluate wavelength dependent arguments
                for param_name in wavelength_dependent_arguments:
                    if not callable(element_parameters[param_name]):
                        # Argument is not callable, so no evaluation can be done.
                        continue

                    if param_name in grid_dependent_arguments:
                        # Argument already handled above.
                        continue

                    # Argument is a function of wavelength.
                    try:
                        element_parameters[param_name] = element_parameters[param_name](wavelength)
                    except Exception:
                        # Function evaluation failed. Raise exception.
                        raise RuntimeError('The argument %s can not be evaluated.' % param_name)

                # Create element.
                elem = optical_element_class(**element_parameters)

                # Add element to cache.
                self._cache[cache_key] = elem

                if grid_dependent:
                    cache_key_output = ('output', elem.output_grid)
                    if wavelength_dependent:
                        cache_key_output += (wavelength_key, )

                    self._cache[cache_key_output] = elem

                return elem

            def forward(self, wavefront, *args, **kwargs):
                return self.get_instance(input_grid=wavefront.electric_field.grid, wavelength=wavefront.wavelength).forward(wavefront, *args, **kwargs)

            def backward(self, wavefront, *args, **kwargs):
                return self.get_instance(output_grid=wavefront.electric_field.grid, wavelength=wavefront.wavelength).backward(wavefront, *args, **kwargs)

            def get_transformation_matrix_forward(self, input_grid, wavelength, *args, **kwargs):
                return self.get_instance(input_grid=input_grid, wavelength=wavelength).get_transformation_matrix_forward(input_grid, wavelength, *args, **kwargs)

            def get_transformation_matrix_backward(self, output_grid, wavelength, *args, **kwargs):
                return self.get_instance(output_grid=output_grid, wavelength=wavelength).get_transformation_matrix_backward(output_grid, wavelength, *args, **kwargs)

        return AgnosticOpticalElement
    return decorator

class OpticalSystem(OpticalElement):
    '''An linear path of optical elements.

    Parameters
    ----------
    optical_elements : list of OpticalElement
        The optical elements in the order that the wavefront propagates.
    '''
    def __init__(self, optical_elements):
        self.optical_elements = optical_elements

    def forward(self, wavefront):
        '''Propagate a wavefront forward through the optical system.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        wf = wavefront

        for optical_element in self.optical_elements:
            wf = optical_element.forward(wf)

        return wf

    def backward(self, wavefront):
        '''Propagate a wavefront backward through the optical system.

        This will be implemented by the derived class.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The propagated wavefront.
        '''
        wf = wavefront

        for optical_element in reversed(self.optical_elements):
            wf = optical_element.backward(wf)

        return wf

    def get_transformation_matrix_forward(self, wavelength=1):
        '''Calculate the forward linear transformation matrix.

        Parameters
        ----------
        wavelength : scalar
            The wavelength for which the transformation matrix will be calculated.

        Returns
        -------
        ndarray
            The full propagation matrix.
        '''
        matrix = np.array(1)

        for optical_element in self.optical_elements:
            matrix = np.dot(optical_element.get_transformation_matrix_forward(wavelength), matrix)

        return matrix

    def get_transformation_matrix_backward(self, wavelength=1):
        '''Calculate the forward linear transformation matrix.

        Parameters
        ----------
        wavelength : scalar
            The wavelength for which the transformation matrix will be calculated.

        Returns
        -------
        ndarray
            The full propagation matrix.
        '''
        matrix = np.array(1)

        for optical_element in reversed(self.optical_elements):
            matrix = np.dot(optical_element.get_transformation_matrix_backward(wavelength), matrix)

        return matrix

    @property
    def optical_elements(self):
        '''The list of optical elements contained in this optical system.
        '''
        return self._optical_elements

    @optical_elements.setter
    def optical_elements(self, optical_elements):
        self._optical_elements = list(optical_elements)



import numpy as np

from .fast_fourier_transform import FastFourierTransform
from ..field import Field, field_dot, field_conjugate_transpose
from .._math import fft as _fft_module


class FourierFilter(object):
    '''A filter in the Fourier domain.

    The filtering is performed by Fast Fourier Transforms, but is quicker than
    the equivalent multiplication in the Fourier domain using the FastFourierTransform
    classes. It does this by avoiding redundant field multiplications that limit performance.

    Parameters
    ----------
    input_grid : Grid
        The grid that is expected for the input field.
    transfer_function : Field generator or Field
        The transfer function to use for the filter. If this is a Field, the user is responsible
        for checking that its grid corresponds to the internal grid used by this filter. The Grid
        is not checked.
    q : scalar
        The amount of zeropadding to perform in the real domain. A value
        of 1 denotes no zeropadding. Zeropadding increases the resolution in the
        Fourier domain and therefore reduces aliasing/wrapping effects.
    '''
    def __init__(self, input_grid, transfer_function, q=1):
        fft = FastFourierTransform(input_grid, q)

        self.input_grid = input_grid
        self.internal_grid = fft.output_grid
        self.cutout = fft.cutout_input
        self.shape_in = input_grid.shape

        self.transfer_function = transfer_function

        self._transfer_function = None
        self.internal_array = None

    def _compute_functions(self, field):
        if self._transfer_function is None or self._transfer_function.dtype != field.dtype:
            if hasattr(self.transfer_function, '__call__'):
                tf = self.transfer_function(self.internal_grid)
            else:
                tf = self.transfer_function.copy()

            tf = np.fft.ifftshift(tf.shaped, axes=tuple(range(-self.input_grid.ndim, 0)))
            self._transfer_function = tf.astype(field.dtype, copy=False)

        recompute_internal_array = self.internal_array is None
        recompute_internal_array = recompute_internal_array or (self.internal_array.ndim != (field.grid.ndim + field.tensor_order))
        recompute_internal_array = recompute_internal_array or (self.internal_array.dtype != field.dtype)
        recompute_internal_array = recompute_internal_array or not np.array_equal(self.internal_array.shape[:field.tensor_order], field.tensor_shape)

        if recompute_internal_array:
            self.internal_array = self.internal_grid.zeros(field.tensor_shape, field.dtype).shaped

    def forward(self, field):
        '''Return the forward filtering of the input field.

        Parameters
        ----------
        field : Field
            The field to filter.

        Returns
        -------
        Field
            The filtered field.
        '''
        return self._operation(field, adjoint=False)

    def backward(self, field):
        '''Return the backward (adjoint) filtering of the input field.

        Parameters
        ----------
        field : Field
            The field to filter.

        Returns
        -------
        Field
            The adjoint filtered field.
        '''
        return self._operation(field, adjoint=True)

    def _operation(self, field, adjoint):
        '''The internal filtering operation.

        Parameters
        ----------
        field : Field
            The input field.
        adjoint : boolean
            Whether to perform a forward or adjoint filter.

        Returns
        -------
        Field
            The filtered field.
        '''
        self._compute_functions(field)

        if self.cutout is None:
            f = field.shaped
        else:
            f = self.internal_array
            f[:] = 0
            c = tuple([slice(None)] * field.tensor_order) + self.cutout
            f[c] = field.shaped

        # Don't overwrite f if it shares memory with the input field.
        overwrite_x = self.cutout is not None
        axes = tuple(range(-self.input_grid.ndim, 0))

        f = _fft_module.fftn(f, axes=axes, overwrite_x=overwrite_x)

        if (self._transfer_function.ndim - self.internal_grid.ndim) == 2:
            # The transfer function is a matrix field.
            s1 = f.shape[:-self.internal_grid.ndim] + (self.internal_grid.size,)
            f = Field(f.reshape(s1), self.internal_grid)

            s2 = self._transfer_function.shape[:-self.internal_grid.ndim] + (self.internal_grid.size,)
            tf = Field(self._transfer_function.reshape(s2), self.internal_grid)

            if adjoint:
                tf = field_conjugate_transpose(tf)

            f = field_dot(tf, f).shaped
        else:
            # The transfer function is a scalar field.
            if adjoint:
                tf = self._transfer_function.conj()
            else:
                tf = self._transfer_function

            # This is faster than f *= tf for Numpy due to it going back to C ordering.
            f = f * tf

        # Since f is now guaranteed to not share memory, always allow overwriting.
        overwrite_x = True

        f = _fft_module.ifftn(f, axes=axes, overwrite_x=overwrite_x)

        s = f.shape[:-self.internal_grid.ndim] + (-1,)
        if self.cutout is None:
            res = f.reshape(s)
        else:
            res = f[c].reshape(s)

        return Field(res, self.input_grid)



import numpy as np

from ..optics import Wavefront, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward
from ..field import Field
from ..fourier import make_fourier_transform

class FraunhoferPropagator(AgnosticOpticalElement):
    '''A monochromatic perfect lens propagator.

        This implements the propagation of a wavefront through a perfect lens. The wavefront
        is assumed to be exactly in the front focal plane of the lens and is propagated to the
        back focal plane. The implementation follows [Goodman2005]_.

        .. [Goodman2005] Goodman, J.W., 2005 Introduction to Fourier optics. Roberts and Company Publishers.

        Parameters
        ----------
        input_grid : Grid
            The grid on which the incoming wavefront is defined.
        output_grid : Grid
            The grid on which the outgoing wavefront is to be evaluated.
        focal_length : scalar
            The focal length of the lens system.
    '''
    def __init__(self, input_grid, output_grid, focal_length=1):
        self._input_grid = input_grid
        self._output_grid = output_grid
        self._focal_length = focal_length

        AgnosticOpticalElement.__init__(self, grid_dependent=True, wavelength_dependent=True)

    def make_instance(self, instance_data, input_grid, output_grid, wavelength):
        focal_length = self.evaluate_parameter(self.focal_length, input_grid, output_grid, wavelength)

        instance_data.uv_grid = output_grid.scaled(2 * np.pi / (focal_length * wavelength))
        instance_data.fourier_transform = make_fourier_transform(input_grid, instance_data.uv_grid)

        instance_data.norm_factor = 1 / (1j * focal_length * wavelength)

    @property
    def focal_length(self):
        return self._focal_length

    @focal_length.setter
    def focal_length(self, focal_length):
        self._focal_length = focal_length

        self.clear_cache()

    def get_input_grid(self, output_grid, wavelength):
        return self._input_grid

    def get_output_grid(self, input_grid, wavelength):
        return self._output_grid

    @make_agnostic_forward
    def forward(self, instance_data, wavefront):
        '''Propagate a wavefront forward through the lens.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The wavefront after the propagation.
        '''
        U_new = instance_data.fourier_transform.forward(wavefront.electric_field) * instance_data.norm_factor
        return Wavefront(Field(U_new, instance_data.output_grid), wavefront.wavelength, wavefront.input_stokes_vector)

    @make_agnostic_backward
    def backward(self, instance_data, wavefront):
        '''Propagate a wavefront backward through the lens.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The wavefront after the propagation.
        '''
        U_new = instance_data.fourier_transform.backward(wavefront.electric_field) / instance_data.norm_factor
        return Wavefront(Field(U_new, instance_data.input_grid), wavefront.wavelength, wavefront.input_stokes_vector)



import numpy as np
from scipy.special import jv, kn
from scipy.optimize import fsolve

from ..mode_basis import ModeBasis

def eigenvalue_equation(u, m, V):
    '''Evaluates the eigenvalue equation for a circular step-index fiber.

    Parameters
    ----------
    u : scalar
        The normalized propagation constant.
    m : int
        The azimuthal order
    V : scalar
        The normalized frequency parameter of the fiber.

    Returns
    -------
    scalar
        The eigenvalue equation value
    '''
    w = np.sqrt(V**2 - u**2)
    return jv(m, u) / (u * jv(m + 1, u)) - kn(m, w) / (w * kn(m + 1, w))

def find_branch_cuts(m, V):
    '''Find all the solutions for the eigenvalue function.

    Parameters
    ----------
    m : int
        The azimuthal order
    V : scalar
        The normalized frequency parameter of the fiber.

    Returns
    -------
    Tuple
        A tuple containing the solutions of the eigenvalue function. If no solutions were found returns None.
    '''
    # Make an initial rough grid
    num_steps = 501
    theta = np.linspace(np.pi * 0.499, 0, num_steps, endpoint=False)
    u = V * np.cos(theta)

    # Find the position where the eigenvalue equation goes through zero
    diff = eigenvalue_equation(u, m, V)
    fu = np.diff(np.sign(diff)) < 0
    ind = np.where(abs(fu - 1) <= 0.01)[0]

    if len(ind) > 0:
        # Refine the zero with a rootfinding algorithm
        u0 = fsolve(eigenvalue_equation, u[ind], args=(m, V))
        w0 = np.sqrt(V**2 - u0**2)
        return u0, w0
    else:
        return None

def lp_radial(m, u, w, r):
    '''Evaluates the radial profile of the LP modes.

    Parameters
    ----------
    m : int
        The azimuthal order
    u : scalar
        The normalized inner propagation constant.
    w : scalar
        The normalized outer propagation constant.
    r : array_like
        The radial coordinates on which to evaluate the bessel modes.

    Returns
    -------
    array_like
        An array that contains the radial profile.
    '''
    # The scaling factor for the continuity condition
    scaling_factor = jv(m, u) / kn(m, w)

    # Find the grid inside and outside the core radius
    mask = r < 1

    # Evaluate the radial mode profile
    mode_field = np.zeros_like(r)
    mode_field[mask] = jv(m, u * r[mask])
    mode_field[~mask] = scaling_factor * kn(m, w * r[~mask])

    return mode_field

def lp_azimuthal(m, theta):
    '''Evaluates the azimuthal profile of the LP modes.

    Parameters
    ----------
    m : int
        The azimuthal order
    theta : array_like
        The azimuthal coordinates on which to evaluate the cosine and sine modes.

    Returns
    -------
    array_like
        An array that contains the azimuthal profile.
    '''
    if m >= 0:
        return np.cos(m * theta)
    else:
        return np.sin(m * theta)

def make_lp_modes(grid, V_number, core_radius, return_betas=False):
    '''Make a ModeBasis out of the guided modes that are supported by a step-index fiber.

    This function solves the eigenvalue equation of for a radial step-index profile fiber.
    It calculates the mode profiles and can also return the propagation constants.

    Parameters
    ----------
    grid : Grid
        The grid on which to calculate the mode basis.
    V_number : scalar
        The normalized frequency parameter of the fiber.
    core_radius : scalar
        The core radius of a step-index fiber.
    return_betas : bool
        If true this function also returns the propagation constants.
    Returns
    -------
    ModeBasis
        An ModeBasis containing all supported LP modes.
    array_like
        An array containing all propagation constants. This is only returned if return_betas is True.
    '''
    finding_new_modes = True
    m = 0
    num_modes = 0

    # scaled grid
    R, Theta = grid.scaled(1 / core_radius).as_('polar').coords
    modes = []
    betas = []
    while finding_new_modes:

        solutions = find_branch_cuts(m, V_number)
        if solutions is not None:
            for ui, wi in zip(solutions[0], solutions[1]):

                radial_prodile = lp_radial(m, ui, wi, R)
                beta = np.sqrt((V_number**2 + wi**2 - ui**2) / (2 * core_radius**2))

                ms = [m, -m] if m > 0 else [m]
                for mi in ms:
                    azimutal_profile = lp_azimuthal(mi, Theta)
                    mode_profile = radial_prodile * azimutal_profile

                    # Normalize the mode numerically as there is no analytical normalization
                    norm = np.sqrt(np.sum(mode_profile * mode_profile.conj() * grid.weights))
                    mode_profile /= norm
                    modes.append(mode_profile)
                    betas.append(beta)
                    num_modes += 1

            m += 1
        else:
            finding_new_modes = False

    # Sort the modes according to their propagation constant
    betas = np.array(betas)
    index_sorting = np.argsort(betas)[::-1]
    modes = np.array(modes)[index_sorting].T
    betas = betas[index_sorting]

    if return_betas:
        return ModeBasis(modes, grid), np.array(betas)
    else:
        return ModeBasis(modes, grid)

def make_LP_modes(*args, **kwargs):  # noqa: N802
    import warnings
    warnings.warn('Use the lower-case version "make_lp_modes" instead.', DeprecationWarning, stacklevel=2)

    return make_lp_modes(*args, **kwargs)



import numpy as np
import copy
import warnings
import xxhash

class Grid(object):
    '''A set of points on some coordinate system.

    Parameters
    ----------
    coords : Coords
        The actual definition of the coordinate values.
    weights : array_like or None
        The interval size, area, volume or hypervolume of each point, depending on the number of dimensions.
        If this is None (default), the weights will be attempted to be calculated on the fly when needed.

    Attributes
    ----------
    coords
        The coordinate values for each dimension.
    '''

    _coordinate_system = 'none'
    _coordinate_system_transformations = {}
    _coordinate_systems = {}

    def __init__(self, coords, weights=None):
        self.coords = coords
        self.weights = weights

    def copy(self):
        '''Create a copy.
        '''
        return copy.deepcopy(self)

    @classmethod
    def from_dict(cls, tree):
        '''Make a Grid from a dictionary, previously created by `to_dict()`.

        Parameters
        ----------
        tree : dictionary
            The dictionary from which to make a new Grid object.

        Returns
        -------
        Grid
            The created object.

        Raises
        ------
        ValueError
            If the dictionary is not formatted correctly.
        '''
        from .coordinates import Coords

        coords = Coords.from_dict(tree['coords'])
        grid_class = Grid._coordinate_systems[tree['coordinate_system']]

        if 'weights' in tree:
            weights = copy.deepcopy(tree['weights'])
        else:
            weights = None

        return grid_class(coords, weights)

    def to_dict(self):
        '''Convert the object to a dictionary for serialization.

        Returns
        -------
        dictionary
            The created dictionary.
        '''
        tree = {
            'coordinate_system': self._coordinate_system,
            'coords': self.coords.to_dict(),
            'weights': self._weights
        }

        return tree

    def subset(self, criterium):
        '''Construct a subset of the current sampling, based on `criterium`.

        Parameters
        ----------
        criterium : function or array_like
            The criterium used to select points. A function will be evaluated for every point.
            Otherwise, this must be a boolean array of integer array, used for slicing the points.

        Returns
        -------
        Grid
            A new grid with UnstructuredCoords that includes only the points for which the criterium
            was true.
        '''
        from .coordinates import UnstructuredCoords

        if hasattr(criterium, '__call__'):
            indices = criterium(self) != 0
        else:
            indices = criterium

        new_coords = [c[indices] for c in self.coords]

        if np.isscalar(self.weights):
            new_weights = self.weights
        else:
            new_weights = self.weights[indices]

        return self.__class__(UnstructuredCoords(new_coords), new_weights)

    @property
    def ndim(self):
        '''The number of dimensions.
        '''
        return len(self.coords)

    @property
    def size(self):
        '''The number of points in this grid.
        '''
        return self.coords.size

    def __len__(self):
        '''The number of points in this grid.
        '''
        return self.size

    @property
    def dims(self):
        '''The number of elements in each dimension for a separated grid.

        Raises
        ------
        ValueError
            If the grid is not separated.
        '''
        if not self.is_separated:
            raise ValueError('A non-separated grid does not have dims.')
        return self.coords.dims

    @property
    def shape(self):
        '''The shape of a reshaped ``numpy.ndarray`` using this grid.

        Raises
        ------
        ValueError
            If the grid is not separated.
        '''
        if not self.is_separated:
            raise ValueError('A non-separated grid does not have a shape.')
        return self.coords.shape

    @property
    def delta(self):
        '''The spacing between points in regularly-spaced grid.

        Raises
        ------
        ValueError
            If the grid is not regular.
        '''
        if not self.is_regular:
            raise ValueError('A non-regular grid does not have a delta.')
        return self.coords.delta

    @property
    def zero(self):
        '''The zero point of a regularly-spaced grid.

        Raises
        ------
        ValueError
            If the grid is not regular.
        '''
        if not self.is_regular:
            raise ValueError('A non-regular grid does not have a zero.')
        return self.coords.zero

    @property
    def separated_coords(self):
        '''A list of coordinates for each dimension in a separated grid.

        Raises
        ------
        ValueError
            If the grid is not separated.
        '''
        if not self.is_separated:
            raise ValueError('A non-separated grid does not have separated coordinates.')
        return self.coords.separated_coords

    @property
    def regular_coords(self):
        '''The tuple (delta, dims, zero) for a regularly-spaced grid.

        Raises
        ------
        ValueError
            If the grid is not regular.
        '''
        if not self.is_regular:
            raise ValueError('A non-regular grid does not have regular coordinates.')
        return self.coords.regular_coords

    @property
    def weights(self):
        '''The interval size, area, volume or hypervolume of each point, depending on the number of dimensions.

        The weights are attempted to be calculated on the fly if not set. If this fails, a warning is emitted and all points will be given an equal weight of one.
        '''
        if self._weights is None:
            self._weights = self.__class__._get_automatic_weights(self.coords)

            if self._weights is None:
                self._weights = 1
                warnings.warn('No automatic weights could be calculated for this grid.', stacklevel=2)

        return self._weights

    @weights.setter
    def weights(self, weights):
        self._weights = weights

    @property
    def points(self):
        '''A list of points of this grid.

        This can be used for easier iteration::

            for p in grid.points:
                print(p)
        '''
        return np.array(self.coords).T

    @property
    def is_separated(self):
        '''True if the grid is separated, False otherwise.
        '''
        return self.coords.is_separated

    @property
    def is_regular(self):
        '''True if the grid is regularly-spaced, False otherwise.
        '''
        return self.coords.is_regular

    @property
    def is_unstructured(self):
        '''True if the grid is unstructured, False otherwise.
        '''
        return self.coords.is_unstructured

    def is_(self, system):
        '''Check if the coordinate system is `system`.

        Parameters
        ----------
        system : str
            The name of the coordinate system to check for.

        Returns
        -------
        bool
            If the coordinate system of the grid is equal to `system`.
        '''
        return system == self._coordinate_system

    def as_(self, system):
        '''Convert the grid to the new coordinate system `system`.

        If the grid is already in the right coordinate system, this function doesn't do anything.

        Parameters
        ----------
        system : str
            The name of the coordinate system to check for.

        Returns
        -------
        Grid
            A new :class:`Grid` in the required coordinate system.

        Raises
        ------
        ValueError
            If the conversion to the coordinate system `system` isn't known.
        '''
        if self.is_(system):
            return self
        else:
            return Grid._coordinate_system_transformations[self._coordinate_system][system](self)

    @staticmethod
    def _add_coordinate_system_transformation(source, dest, func):
        if source in Grid._coordinate_system_transformations:
            Grid._coordinate_system_transformations[source][dest] = func
        else:
            Grid._coordinate_system_transformations[source] = {dest: func}

    @staticmethod
    def _add_coordinate_system(name, grid_class):
        Grid._coordinate_systems[name] = grid_class

    def __getitem__(self, i):
        '''The `i`-th point in this grid.
        '''
        return self.points[i]

    def scale(self, scale):
        '''Scale the grid in-place.

        Parameters
        ----------
        scale : array_like
            The factor with which to scale the grid.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        raise NotImplementedError()

    def scaled(self, scale):
        '''A scaled copy of this grid.

        Parameters
        ----------
        scale : array_like
            The factor with which to scale the grid.

        Returns
        -------
        Grid
            The scaled grid.
        '''
        grid = self.copy()
        grid.scale(scale)
        return grid

    def shift(self, shift):
        '''Shift the grid in-place.

        Parameters
        ----------
        shift : array_like
            The amount with which to shift the grid.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        raise NotImplementedError()

    def shifted(self, shift):
        '''A shifted copy of this grid.

        Parameters
        ----------
        shift : array_like
            The amount with which to shift the grid.

        Returns
        -------
        Grid
            The scaled grid.
        '''
        grid = self.copy()
        grid.shift(shift)
        return grid

    def rotate(self, angle, axis=None):
        '''Rotate the grid in-place.

        Parameters
        ----------
        angle : scalar
            The angle in radians.
        axis : ndarray or None
            The axis of rotation. For two-dimensional grids, it is ignored. For
            three-dimensional grids it is required.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        raise NotImplementedError()

    def rotated(self, angle, axis=None):
        '''A rotated copy of this grid.

        Parameters
        ----------
        angle : scalar
            The angle in radians.
        axis : ndarray or None
            The axis of rotation. For two-dimensional grids, it is ignored. For
            three-dimensional grids it is required.

        Returns
        -------
        Grid
            The rotated grid.
        '''
        grid = self.copy()
        grid.rotate(angle, axis)
        return grid

    def reverse(self):
        '''Reverse the order of the points in-place.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        self.coords.reverse()
        return self

    def reversed(self):
        '''Make a copy of the grid with the order of the points reversed.

        Returns
        -------
        Grid
            The reversed grid.
        '''
        grid = self.copy()
        grid.reverse()
        return grid

    @staticmethod
    def _get_automatic_weights(coords):
        raise NotImplementedError()

    def __str__(self):
        return str(self.__class__.__name__) + '(' + str(self.coords.__class__.__name__) + ')'

    def __hash__(self):
        h = xxhash.xxh64()
        h.update(self._coordinate_system)

        if self.is_regular:
            h.update(self.delta)
            h.update(self.dims)
            h.update(self.zero)
        elif self.is_separated:
            for s in self.separated_coords:
                h.update(s)
        else:
            for s in self.coords:
                h.update(s)

        return h.intdigest()

    def __eq__(self, other):
        '''Check equality of two grids.

        .. note::
            This will return False if the two grids have different coordinate systems or coordinate types,
            even if the underlying points are identical

        Parameters
        ----------
        other : object
            The object to which to compare to.

        Returns
        -------
        boolean
            Whether the two objects are identical.
        '''
        if not isinstance(other, Grid):
            return False

        if self._coordinate_system != other._coordinate_system:
            return False

        return self.coords == other.coords

    def closest_to(self, p):
        '''Get the index of the point closest to point `p`.

        Point `p` is assumed to have the same coordinate system as the grid itself.

        Parameters
        ----------
        p : array_like
            The point at which to search for.

        Returns
        -------
        int
            The index of the closest point.
        '''
        rel_points = self.points - np.array(p) * np.ones(self.ndim)
        return np.argmin(np.sum(rel_points**2, axis=-1))

    def zeros(self, tensor_shape=None, dtype=None):
        '''Create a field of zeros from this `Grid`.

        Parameters
        ----------
        tensor_shape : array_like or None
            The shape of the tensors in the to be created field. If this is None,
            a scalar field will be created.
        dtype : data-type
            The numpy data-type with which to create the field.

        Returns
        -------
        Field
            A zeros field.
        '''
        from .field import Field

        if tensor_shape is None:
            shape = [self.size]
        else:
            shape = np.concatenate((tensor_shape, [self.size]))

        return Field(np.zeros(shape, dtype), self)

    def ones(self, tensor_shape=None, dtype=None):
        '''Create a field of ones from this `Grid`.

        Parameters
        ----------
        tensor_shape : array_like or None
            The shape of the tensors in the to be created field. If this is None,
            a scalar field will be created.
        dtype : data-type
            The numpy data-type with which to create the field.

        Returns
        -------
        Field
            A ones field.
        '''
        from .field import Field

        if tensor_shape is None:
            shape = [self.size]
        else:
            shape = np.concatenate((tensor_shape, [self.size]))

        return Field(np.ones(shape, dtype=dtype), self)

    def empty(self, tensor_shape=None, dtype=None):
        '''Create an empty Field from this `Grid`.

        Parameters
        ----------
        tensor_shape : array_like or None
            The shape of the tensors in the to be created field. If this is None,
            a scalar field will be created.
        dtype : data-type
            The numpy data-type with which to create the field.

        Returns
        -------
        Field
            A empty field.
        '''
        from .field import Field

        if tensor_shape is None:
            shape = [self.size]
        else:
            shape = np.concatenate((tensor_shape, [self.size]))

        return Field(np.empty(shape, dtype=dtype), self)



import numpy as np
from ..field import Field

def large_poisson(lam, thresh=1e6):
    """
    Draw samples from a Poisson distribution, taking care of large values of `lam`.

    At large values of `lam` the distribution automatically switches to the corresponding normal distribution.
    This switch is independently decided for each expectation value in the `lam` array.

    Parameters
    ----------
    lam : array_like
        Expectation value for the Poisson distribution. Must be >= 0.
    thresh : float
        The threshold at which the distribution switched from a Poisson to a normal distribution.

    Returns
    -------
    array_like
        The drawn samples from the Poisson or normal distribution, depending on the expectation value.
    """
    large = lam > thresh
    small = ~large

    # Use normal approximation if the number of photons is large
    n = np.zeros(lam.shape)
    n[large] = np.round(lam[large] + np.random.normal(size=np.sum(large)) * np.sqrt(lam[large]))
    n[small] = np.random.poisson(lam[small], size=np.sum(small))

    if hasattr(lam, 'grid'):
        n = Field(n, lam.grid)

    return n

def large_gamma(lam, theta, thresh=1e6):
    """
    Draw samples from a Gamma distribution, taking care of large values of `lam`.

    At large values of `lam` the distribution automatically switches to the corresponding normal distribution.
    This switch is independently decided for each expectation value in the `lam` array.

    Parameters
    ----------
    lam : array_like
        The shape parameter distribution. Must be >= 0.
    theta : array_like
        The scale parameter of the Gamma distribution.
    thresh : float
        The threshold at which the distribution switched from a Gamma to a normal distribution.

    Returns
    -------
    array_like
        The drawn samples from the Gamma or normal distribution, depending on the expectation value.
    """
    large = lam > thresh
    small = ~large

    theta = theta * np.ones_like(lam)

    # Use normal approximation if the shape parameter is large
    n = np.zeros(lam.shape)
    mean = lam * theta
    std = np.sqrt(mean * theta)

    n[large] = mean[large] + std[large] * np.random.normal(size=np.sum(large))
    n[small] = np.random.gamma(lam[small], theta[small], size=np.sum(small))

    if hasattr(lam, 'grid'):
        n = Field(n, lam.grid)

    return n

def make_emccd_noise(photo_electron_flux, read_noise, emgain):
    """
    Draw samples from a Poisson-Gamma-Normal distribution, which is an accurate model of EMCCDs (following [Hirsch2013]_).

    The model is most accurate for high emgains.

    .. [Hirsch2013] Michael Hirsch et al. " A stochastic model for electron multiplication
        charge-coupled devices - from theory to practice." PloS one 8.1 (2013): e53671.

    Parameters
    ----------
    photo_electron_flux : array_like
        The incoming photo-electron flux, which is usually a combination of QE * (photon_flux + background) + dark current.
    read_noise : array_like
        The read noise of the camera.
    emgain : float
        The electron multiplying gain of the EMCCD process. Must be larger than 0.

    Returns
    -------
    array_like
        The noisy realization of the EMCCD detection.
    """
    photo_electrons = large_poisson(photo_electron_flux)

    detector_counts = large_gamma(photo_electrons, emgain) + read_noise * np.random.standard_normal(photo_electron_flux.shape)

    return detector_counts
